from typing import List

import numpy as np


def get_class_count(truth_index: List[int],
                    truth: np.ndarray,
                    class_weights_threshold: List[float]) -> List[int]:
    class_count = [0] * len(truth_index)
    for idx, cl in enumerate(truth_index):
        truth_sum = int(np.sum(truth[cl - 1,] >= class_weights_threshold[cl - 1]))
        class_count[idx] = truth_sum

    return class_count


def get_total_class_count(mixdb: dict) -> List[int]:
    total_class_count = [0] * mixdb['num_classes']
    for mixture in mixdb['mixtures']:
        for idx, cl in enumerate(mixdb['targets'][mixture['target_file_index']]['truth_index']):
            total_class_count[cl - 1] += int(mixture['class_count'][idx])

    return total_class_count



from copy import deepcopy

import numpy as np
from pyaaware import ForwardTransform
from pyaaware import SED

from sonusai import logger
from sonusai.mixture import get_class_count
from sonusai.mixture import get_class_weights_threshold
from sonusai.utils import int16_to_float


def strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: dict, mixture_record: dict, target_audio: np.ndarray) -> np.ndarray:
    truth_config = deepcopy(mixdb['targets'][mixture_record['target_file_index']]['truth_config'])
    truth_config['index'] = mixdb['targets'][mixture_record['target_file_index']]['truth_index']
    truth_config['frame_size'] = mixdb['frame_size']
    truth_config['num_classes'] = mixdb['num_classes']
    truth_config['mutex'] = mixdb['truth_mutex']

    if mixture_record['target_gain'] == 0:
        truth = np.zeros((mixdb['num_classes'], len(target_audio)), dtype=np.single)
    else:
        truth = truth_function(
            audio=np.int16(np.single(target_audio) / mixture_record['target_gain']),
            function=mixdb['targets'][mixture_record['target_file_index']]['truth_function'],
            config=truth_config)

    mixture_record['class_count'] = get_class_count(
        truth_index=truth_config['index'],
        truth=truth,
        class_weights_threshold=get_class_weights_threshold(mixdb))

    return truth


def truth_function(audio: np.ndarray, function: str, config: dict) -> np.ndarray:
    required_parameters = ['index', 'frame_size', 'num_classes', 'mutex']
    for parameter in required_parameters:
        if parameter not in config.keys():
            logger.error('Truth function config missing required parameter: {}'.format(parameter))
            exit()

    index = config['index']
    frame_size = config['frame_size']
    num_classes = config['num_classes']
    mutex = config['mutex']

    if function == 'sed':
        if len(audio) % frame_size != 0:
            logger.error('Number of samples in audio is not a multiple of {}'.format(frame_size))
            exit()

        if 'thresholds' in config.keys():
            thresholds = config['thresholds']
            if not isinstance(thresholds, list) or len(thresholds) != 3:
                logger.error('Truth function SED thresholds does not contain 3 entries: {}'.format(thresholds))
                exit()
            if not strictly_decreasing(thresholds):
                logger.error('Truth function SED thresholds are not strictly decreasing: {}'.format(thresholds))
                exit()
        else:
            thresholds = None

        fft = ForwardTransform(N=frame_size * 4, R=frame_size)
        sed = SED(thresholds=thresholds,
                  index=index,
                  frame_size=frame_size,
                  num_classes=num_classes,
                  mutex=mutex)

        truth = np.empty((num_classes, 0), dtype=np.single)
        for offset in range(0, len(audio), frame_size):
            new_truth = sed.execute(fft.energy(int16_to_float(audio[offset:offset + frame_size])))
            truth = np.hstack((truth, np.reshape(new_truth, (len(new_truth), 1))))

        # This repeats each sample and not the whole sequence
        truth = truth.repeat(frame_size, axis=1)
        return truth

    logger.error('Unsupported truth function: {}'.format(function))
    exit()



"""sonusai

usage: sonusai [--version] [--help] <command> [<args>...]

The sonusai commands are:
   evaluate     Evaluate model performance
   genft        Generate feature and truth data
   genmix       Generate mixture and truth data
   genmixdb     Generate a mixture database
   gentcst      Generate target configuration from a subdirectory tree
   mkwav        Make WAV files from training data
   predict      Run predict on a trained model

Aaware Sound and Voice Machine Learning Framework. See 'sonusai help <command>'
for more information on a specific command.

"""
from subprocess import call

from docopt import docopt

import sonusai
from sonusai import logger
from sonusai.utils import trim_docstring


def main():
    try:
        command_list = 'evaluate genft genmix genmixdb gentcst mkwav predict'.split()

        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        command = args['<command>']
        argv = args['<args>']

        if command is None:
            exit(call([__file__, '--help']))
        elif command == 'help':
            if not argv:
                exit(call([__file__, '--help']))
            elif argv[0] in command_list:
                exit(call(['sonusai-{}'.format(argv[0]), '-h']))
            else:
                logger.error("{} is not a SonusAI command. See 'sonusai help'.".format(argv[0]))
                exit()
        elif command in command_list:
            exit(call(['sonusai-{}'.format(command), *argv]))

        logger.error("{} is not a SonusAI command. See 'sonusai help'.".format(command))
        exit()

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()


if __name__ == '__main__':
    main()



import json
from os.path import exists

from sonusai import logger


def load_mixdb(name: str) -> dict:
    if not exists(name):
        logger.error('{} does not exist'.format(name))
        exit()

    with open(name, encoding='utf-8') as f:
        mixdb = json.load(f)

    return mixdb



import h5py
import numpy as np

from sonusai import logger


def read_predict_data(filename: str, expected_frames: int) -> (None, np.ndarray):
    if not filename:
        return None

    logger.info('Reading prediction data from {}'.format(filename))
    with h5py.File(name=filename, mode='r') as f:
        # prediction data is either [frames, num_classes], or [frames, timesteps, num_classes]
        predict = np.array(f['predict'])

        if predict.ndim == 2:
            frames, num_classes = predict.shape

            if frames != expected_frames:
                logger.warning('Ignoring prediction data in {} due to frames mismatch'.format(filename))
                return None
        elif predict.ndim == 3:
            frames, timesteps, num_classes = predict.shape

            if frames * timesteps != expected_frames:
                logger.warning('Ignoring prediction data in {} due to frames mismatch'.format(filename))
                return None

            logger.info('Reshaping prediction data in {} from [{}, {}, {}] to [{}, {}]'.
                        format(filename, frames, timesteps, num_classes, frames * timesteps, num_classes))
            predict = np.reshape(predict, [frames * timesteps, num_classes], order='F')
        else:
            logger.warning('Ignoring prediction data in {} due to invalid dimensions'.format(filename))
            return None

        return predict



import random
from typing import List
from typing import Union

import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai import logger
from sonusai.genft import genft
from sonusai.mixture import get_mixtures_from_mixid
from sonusai.utils import reshape_inputs


class DataGenerator(Sequence):
    """Generates data for Keras"""

    def __init__(self,
                 mixdb: dict,
                 mixid: Union[str, List[int]],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False,
                 chunks: int = 8):
        """Initialization"""
        self.mixdb = mixdb
        self.mixid = mixid
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle
        self.chunks = chunks

        self.mixtures = None
        self.index_map = None
        self.batches_per_epoch = None
        self.features_per_batch = None
        self.feature = None
        self.truth = None
        self.cached_indices = None
        self.cache_misses = []

        self.initialize_mixtures()

    def __len__(self):
        """Denotes the number of batches per epoch"""
        return self.batches_per_epoch

    def __getitem__(self, index):
        """Generate one batch of data"""
        if self.cached_indices is None or index < self.cached_indices[0] or index > self.cached_indices[1]:
            self.fetch_data_for_index(index)

        offset = index - self.cached_indices[0]
        feature, truth, _, _, _, _ = reshape_inputs(feature=self.feature[offset],
                                                    truth=self.truth[offset],
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self):
        """Modification of dataset between epochs"""
        if self.shuffle:
            random.shuffle(self.mixid)
            self.initialize_mixtures()

    def get_feature_frames(self, mixtures: list, mixid: Union[str, List[int]]) -> int:
        subset = get_mixtures_from_mixid(mixtures, mixid)
        return sum([sub['samples'] for sub in subset]) // self.mixdb['feature_step_samples']

    def initialize_mixtures(self):
        self.mixtures = get_mixtures_from_mixid(self.mixdb['mixtures'], self.mixid)

        feature_frames = self.get_feature_frames(self.mixtures, ':')
        self.features_per_batch = self.batch_size if self.timesteps == 0 else self.batch_size * self.timesteps
        self.batches_per_epoch = feature_frames // self.features_per_batch

        if self.batches_per_epoch == 0:
            logger.error(
                'Error: dataset only contains {} frames which is not enough to fill a batch size of {}. '
                'Either provide more data or decrease the batch size.'.format(
                    feature_frames, self.features_per_batch))
            exit()

        # Compute mixid and offset for dataset
        # offsets are needed because mixtures are not guaranteed to fall on batch boundaries
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence.
        offsets = np.cumsum([sub['samples'] for sub in self.mixtures]) // self.mixdb['feature_step_samples'] - 1
        self.index_map = [{} for n in range(self.batches_per_epoch)]
        cur_offset = 0
        prv_id = 0
        for n in range(self.batches_per_epoch):
            mixid = []
            offset = []
            for index in range(n * self.features_per_batch, (n + 1) * self.features_per_batch):
                cur_id = next(idx for idx, val in enumerate(offsets) if val >= index)
                if cur_id != prv_id:
                    cur_offset = 0
                offset.append(cur_offset)
                cur_offset += 1
                prv_id = cur_id
                mixid.append(cur_id)
            mixid = sorted(list(set(mixid)))

            self.index_map[n]['mixid'] = mixid
            self.index_map[n]['offset'] = offset

    def fetch_data_for_index(self, index: int):
        self.cache_misses.append(index)
        mixid = self.index_map[index]['mixid']
        offset = self.index_map[index]['offset'][0]
        # get_feature_frames() does not account for mixture start offset, so the actual number of batches
        # returned from genft() may be less than chunks
        while (self.get_feature_frames(self.mixtures, mixid) // self.features_per_batch) < self.chunks and \
                mixid[-1] < len(self.mixtures) - 1:
            mixid.append(mixid[-1] + 1)

        feature, truth, _, _ = genft(mixdb=self.mixdb,
                                     mixid=mixid,
                                     start_offset=offset,
                                     logging=False)

        features = (feature.shape[0] // self.features_per_batch) * self.features_per_batch
        batches = features // self.features_per_batch

        if batches == 0:
            logger.error(
                'Error: genft returned {} frames which is not enough to fill a batch size of {}. '
                'Either provide more data or decrease the batch size'.format(
                    feature.shape[0], self.features_per_batch))
            exit()

        self.feature = feature[:features].reshape(
            (batches, self.features_per_batch, feature.shape[1], feature.shape[2]))
        self.truth = truth[:features].reshape((batches, self.features_per_batch, truth.shape[1]))

        self.cached_indices = (index, index + batches - 1)

    def get_cache_misses(self):
        return self.cache_misses



"""genft

usage: genft [-hvs] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture IDs JSON file.
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI feature/truth file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A JSON file containing a list of mixture IDs. The list should be named 'mixid'.
                If no file is provided, then all mixtures in the database will be generated.

Outputs:
    OUTPUT.h5   A SonusAI feature HDF5 file. Contains:
                    dataset:    feature
                    dataset:    truth_f
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genft.log

"""

import json
from os.path import splitext
from typing import List
from typing import Union

import h5py
import numpy as np
from docopt import docopt
from pyaaware import FeatureGenerator
from pyaaware import ForwardTransform
from tqdm import tqdm

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import build_noise_audio_db
from sonusai.mixture import build_target_audio_db
from sonusai.mixture import generate_truth
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_mixdb
from sonusai.mixture import load_mixid
from sonusai.mixture import new_mixdb_from_mixid
from sonusai.mixture import process_mixture_audio
from sonusai.mixture import truth_reduction
from sonusai.utils import grouper
from sonusai.utils import human_readable_size
from sonusai.utils import int16_to_float
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring


def genft_init(mixdb: dict,
               mixid: Union[str, List[int]],
               start_offset: int = 0,
               logging: bool = True) -> (dict, int, int, int):
    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub['samples'] for sub in mixdb_out['mixtures']])
    transform_frames = total_samples // mixdb_out['frame_size']
    feature_frames = total_samples // mixdb_out['feature_step_samples'] - start_offset

    if logging:
        logger.info('')
        logger.info('Found {} mixtures to process'.format(len(mixdb_out['mixtures'])))
        logger.info('{} samples, {} transform frames, {} feature frames'.format(total_samples, transform_frames,
                                                                                feature_frames))

    return mixdb_out, total_samples, transform_frames, feature_frames


def genft(mixdb: dict,
          mixid: Union[str, List[int]],
          start_offset: int = 0,
          compute_segsnr: bool = False,
          logging: bool = True,
          show_progress: bool = False,
          progress: tqdm = None) -> (np.ndarray, np.ndarray, dict):
    mixdb_out, total_samples, transform_frames, feature_frames = genft_init(mixdb=mixdb,
                                                                            mixid=mixid,
                                                                            start_offset=start_offset,
                                                                            logging=logging)

    fft = ForwardTransform(N=mixdb_out['frame_size'] * 4, R=mixdb_out['frame_size'])
    fg = FeatureGenerator(frame_size=mixdb_out['frame_size'],
                          feature_mode=mixdb_out['feature'],
                          num_classes=mixdb_out['num_classes'],
                          truth_mutex=mixdb_out['truth_mutex'])

    feature = np.empty((feature_frames, fg.stride, fg.num_bands), dtype=np.single)
    truth_f = np.empty((feature_frames, fg.num_classes), dtype=np.single)

    noise_audios = build_noise_audio_db(mixdb_out)
    target_audios = build_target_audio_db(mixdb_out)

    segsnr = np.empty(0)
    if compute_segsnr:
        segsnr = np.empty(transform_frames, dtype=np.single)

    i_sample_offset = 0
    i_frame_offset = 0
    o_frame_offset = 0
    transform_frame = 0
    feature_frame = 0
    if progress is None:
        progress = tqdm(total=len(mixdb_out['mixtures']), desc='genft', disable=not show_progress)
    for mixture_record in mixdb_out['mixtures']:
        noise_audio, target_audio = process_mixture_audio(i_frame_offset=i_frame_offset,
                                                          i_sample_offset=i_sample_offset,
                                                          o_frame_offset=o_frame_offset,
                                                          mixdb=mixdb_out,
                                                          mixture_record=mixture_record,
                                                          target_audios=target_audios,
                                                          noise_audios=noise_audios)

        mixture_audio = np.array(target_audio + noise_audio, dtype=np.int16)

        truth = generate_truth(mixdb=mixdb_out, mixture_record=mixture_record, target_audio=target_audio)

        for offset in range(0, mixture_record['samples'], mixdb_out['frame_size']):
            if compute_segsnr:
                target_energy = fft.energy(int16_to_float(target_audio[offset:offset + mixdb_out['frame_size']]))
                noise_energy = fft.energy(int16_to_float(noise_audio[offset:offset + mixdb_out['frame_size']]))
                segsnr[transform_frame] = np.single(target_energy / noise_energy)
                transform_frame += 1

            mixture_fd = fft.execute(int16_to_float(mixture_audio[offset:offset + mixdb_out['frame_size']]))
            fg.execute(mixture_fd,
                       truth_reduction(truth[:, offset:offset + mixdb_out['frame_size']],
                                       mixdb_out['truth_reduction_function']))
            if fg.eof():
                # start_offset is used when generating one batch at a time and makes it
                # possible to preserve the mixture data sequence across batch boundaries
                if start_offset != 0:
                    start_offset -= 1
                else:
                    feature[feature_frame] = np.reshape(fg.feature(), (fg.stride, fg.num_bands))
                    truth_f[feature_frame] = fg.truth()
                    feature_frame += 1

        fft.reset()
        fg.reset()

        i_sample_offset += mixture_record['samples']
        i_frame_offset += mixture_record['samples'] // mixdb_out['frame_size']
        o_frame_offset += mixture_record['samples'] // mixdb_out['feature_step_samples']
        progress.update()

    mixdb_out['class_count'] = get_total_class_count(mixdb_out)

    duration = total_samples / sonusai.mixture.sample_rate
    if logging:
        logger.info('')
        logger.info('Duration: {}'.format(seconds_to_hms(seconds=duration)))
        logger.info('feature:  {}'.format(human_readable_size(feature.nbytes, 1)))
        logger.info('truth_f:  {}'.format(human_readable_size(truth_f.nbytes, 1)))
        if compute_segsnr:
            logger.info('segsnr:   {}'.format(human_readable_size(segsnr.nbytes, 1)))

    return feature, truth_f, segsnr, mixdb_out


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        mixdb_name = args['--mixdb']
        mixid_name = args['--mixid']
        output_name = args['--output']
        compute_segsnr = args['--segsnr']

        if not output_name:
            output_name = splitext(mixdb_name)[0] + '.h5'

        log_name = 'genft.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('genft')

        mixdb = load_mixdb(name=mixdb_name)
        mixid = load_mixid(name=mixid_name, mixdb=mixdb)

        mixdb_out, total_samples, _, _ = genft_init(mixdb=mixdb,
                                                    mixid=mixid,
                                                    logging=True)

        chunk_size = 100
        progress = tqdm(total=len(mixid), desc='genft')
        mixid = grouper(range(len(mixdb_out['mixtures'])), chunk_size)
        mixdb_out['class_count'] = [0] * mixdb_out['num_classes']

        i_frame_offset = 0
        o_frame_offset = 0
        chunk_offset = 0
        feature_elems = 0
        truth_elems = 0
        with h5py.File(output_name, 'w') as f:
            for m in mixid:
                feature, truth_f, segsnr, mixdb_tmp = genft(mixdb=mixdb_out,
                                                            mixid=m,
                                                            compute_segsnr=compute_segsnr,
                                                            logging=False,
                                                            progress=progress)
                progress.refresh()

                o_frames = feature.shape[0]
                if o_frames != truth_f.shape[0]:
                    logger.error(
                        'truth_f frames does not match feature frames: {} != {}'.format(o_frames, truth_f.shape[0]))
                    exit()

                if o_frame_offset == 0:
                    feature_dataset = f.create_dataset(name='feature',
                                                       data=feature,
                                                       maxshape=(None, feature.shape[1], feature.shape[2]))
                    truth_dataset = f.create_dataset(name='truth_f',
                                                     data=truth_f,
                                                     maxshape=(None, truth_f.shape[1]))
                    feature_elems = feature.shape[1] * feature.shape[2]
                    truth_elems = truth_f.shape[1]
                else:
                    feature_dataset.resize(o_frame_offset + o_frames, axis=0)
                    feature_dataset[o_frame_offset:] = feature
                    truth_dataset.resize(o_frame_offset + o_frames, axis=0)
                    truth_dataset[o_frame_offset:] = truth_f
                o_frame_offset += o_frames

                if compute_segsnr:
                    i_frames = segsnr.shape[0]
                    if i_frame_offset == 0:
                        segsnr_dataset = f.create_dataset(name='segsnr',
                                                          data=segsnr,
                                                          maxshape=(None,))
                    else:
                        segsnr_dataset.resize(i_frame_offset + i_frames, axis=0)
                        segsnr_dataset[i_frame_offset:] = segsnr
                    i_frame_offset += i_frames

                for idx, val in enumerate(m):
                    mixdb_out['mixtures'][val] = mixdb_tmp['mixtures'][idx]
                for idx in range(mixdb_out['num_classes']):
                    mixdb_out['class_count'][idx] += mixdb_tmp['class_count'][idx]
                chunk_offset += chunk_size
            f.attrs['mixdb'] = json.dumps(mixdb_out)
        progress.close()

        logger.info('Wrote {}'.format(output_name))
        duration = total_samples / sonusai.mixture.sample_rate
        logger.info('')
        logger.info('Duration: {}'.format(seconds_to_hms(seconds=duration)))
        logger.info('feature:  {}'.format(human_readable_size(o_frame_offset * feature_elems * 4, 1)))
        logger.info('truth_f:  {}'.format(human_readable_size(o_frame_offset * truth_elems * 4, 1)))
        if compute_segsnr:
            logger.info('segsnr:   {}'.format(human_readable_size(i_frame_offset * 4, 1)))

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()


if __name__ == '__main__':
    main()



import numpy as np

from sonusai import logger


def truth_reduction(x: np.ndarray, func: str) -> np.ndarray:
    if func == 'max':
        return np.max(x, axis=1)

    if func == 'mean':
        return np.mean(x, axis=1)

    logger.error('Invalid truth reduction function: {}'.format(func))
    exit()



"""genmix

usage: genmix [-hvts] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture IDs JSON file.
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -t, --truth                     Save truth_t. [default: False].
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI mixture file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A JSON file containing a list of mixture IDs. The list should be named 'mixid'.
                If no file is provided, then all mixtures in the database will be generated.

Outputs:
    OUTPUT.h5   A SonusAI mixture HDF5 file. Contains:
                    dataset:    mixture
                    dataset:    truth_t (optional)
                    dataset:    target
                    dataset:    noise
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genmix.log
"""
import json
from os.path import splitext
from typing import List
from typing import Union

import h5py
import numpy as np
from docopt import docopt
from pyaaware import ForwardTransform
from tqdm import tqdm

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import build_noise_audio_db
from sonusai.mixture import build_target_audio_db
from sonusai.mixture import generate_truth
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_mixdb
from sonusai.mixture import load_mixid
from sonusai.mixture import new_mixdb_from_mixid
from sonusai.mixture import process_mixture_audio
from sonusai.utils import grouper
from sonusai.utils import human_readable_size
from sonusai.utils import int16_to_float
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring


def genmix_init(mixdb: dict,
                mixid: Union[str, List[int]],
                logging: bool = True) -> (dict, int):
    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub['samples'] for sub in mixdb_out['mixtures']])

    if logging:
        logger.info('')
        logger.info('Found {} mixtures to process'.format(len(mixdb_out['mixtures'])))
        logger.info('{} samples'.format(total_samples))

    return mixdb_out, total_samples


def genmix(mixdb: dict,
           mixid: Union[str, List[int]],
           compute_segsnr: bool = False,
           save_truth: bool = False,
           logging: bool = False,
           show_progress: bool = False,
           progress: tqdm = None) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, dict):
    mixdb_out, total_samples = genmix_init(mixdb=mixdb, mixid=mixid, logging=logging)

    target = np.empty(total_samples, dtype=np.int16)
    noise = np.empty(total_samples, dtype=np.int16)
    truth_t = np.empty(0)
    if save_truth:
        truth_t = np.empty((mixdb_out['num_classes'], total_samples), dtype=np.single)

    segsnr = np.empty(0)
    fft = ForwardTransform(N=mixdb_out['frame_size'] * 4, R=mixdb_out['frame_size'])
    if compute_segsnr:
        segsnr = np.empty(total_samples, dtype=np.single)

    noise_audios = build_noise_audio_db(mixdb_out)
    target_audios = build_target_audio_db(mixdb_out)

    i_sample_offset = 0
    i_frame_offset = 0
    o_frame_offset = 0
    if progress is None:
        progress = tqdm(total=len(mixdb_out['mixtures']), desc='genmix', disable=not show_progress)
    for mixture_record in mixdb_out['mixtures']:
        noise_audio, target_audio = process_mixture_audio(i_frame_offset=i_frame_offset,
                                                          i_sample_offset=i_sample_offset,
                                                          o_frame_offset=o_frame_offset,
                                                          mixdb=mixdb_out,
                                                          mixture_record=mixture_record,
                                                          target_audios=target_audios,
                                                          noise_audios=noise_audios)

        audio_indices = slice(i_sample_offset, i_sample_offset + len(target_audio))
        target[audio_indices] = target_audio
        noise[audio_indices] = noise_audio

        truth = generate_truth(mixdb=mixdb_out,
                               mixture_record=mixture_record,
                               target_audio=target_audio)
        if save_truth:
            truth_t[:, audio_indices] = truth

        if compute_segsnr:
            for offset in range(0, mixture_record['samples'], mixdb_out['frame_size']):
                target_energy = fft.energy(int16_to_float(target_audio[offset:offset + mixdb_out['frame_size']]))
                noise_energy = fft.energy(int16_to_float(noise_audio[offset:offset + mixdb_out['frame_size']]))
                frame_offset = i_sample_offset + offset
                segsnr[frame_offset:frame_offset + mixdb_out['frame_size']] = np.single(target_energy / noise_energy)

        i_sample_offset += mixture_record['samples']
        i_frame_offset += mixture_record['samples'] // mixdb_out['frame_size']
        o_frame_offset += mixture_record['samples'] // mixdb_out['feature_step_samples']
        progress.update()

    mixture = np.array(target + noise, dtype=np.int16)

    mixdb_out['class_count'] = get_total_class_count(mixdb_out)

    duration = len(mixture) / sonusai.mixture.sample_rate
    if logging:
        logger.info('')
        logger.info('Duration: {}'.format(seconds_to_hms(seconds=duration)))
        logger.info('mixture:  {}'.format(human_readable_size(mixture.nbytes, 1)))
        if save_truth:
            logger.info('truth_t:  {}'.format(human_readable_size(truth_t.nbytes, 1)))
        if compute_segsnr:
            logger.info('segsnr:   {}'.format(human_readable_size(segsnr.nbytes, 1)))

    return mixture, truth_t, target, noise, segsnr, mixdb_out


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        mixdb_name = args['--mixdb']
        mixid_name = args['--mixid']
        output_name = args['--output']
        compute_segsnr = args['--segsnr']
        save_truth = args['--truth']

        if not output_name:
            output_name = splitext(mixdb_name)[0] + '.h5'

        log_name = 'genmix.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('genmix')

        mixdb = load_mixdb(name=mixdb_name)
        mixid = load_mixid(name=mixid_name, mixdb=mixdb)

        mixdb_out, total_samples = genmix_init(mixdb=mixdb,
                                               mixid=mixid,
                                               logging=True)

        chunk_size = 50
        progress = tqdm(total=len(mixid), desc='genmix')
        mixid = grouper(range(len(mixdb_out['mixtures'])), chunk_size)
        mixdb_out['class_count'] = [0] * mixdb_out['num_classes']

        sample_offset = 0
        chunk_offset = 0
        truth_elems = 0
        with h5py.File(output_name, 'w') as f:
            for m in mixid:
                mixture, truth_t, target, noise, segsnr, mixdb_tmp = genmix(mixdb=mixdb_out,
                                                                            mixid=m,
                                                                            compute_segsnr=compute_segsnr,
                                                                            save_truth=save_truth,
                                                                            logging=False,
                                                                            progress=progress)
                progress.refresh()

                samples = mixture.shape[0]
                if save_truth and samples != truth_t.shape[1]:
                    logger.error(
                        'truth_t samples does not match mixture samples: {} != {}'.format(samples, truth_t.shape[1]))
                    exit()
                if samples != target.shape[0]:
                    logger.error(
                        'target samples does not match mixture samples: {} != {}'.format(samples, target.shape[0]))
                    exit()
                if samples != noise.shape[0]:
                    logger.error(
                        'noise samples does not match mixture samples: {} != {}'.format(samples, noise.shape[0]))
                    exit()
                if compute_segsnr and samples != segsnr.shape[0]:
                    logger.error(
                        'segsnr samples does not match mixture samples: {} != {}'.format(samples, segsnr.shape[0]))
                    exit()

                if sample_offset == 0:
                    mixture_dataset = f.create_dataset(name='mixture',
                                                       data=mixture,
                                                       maxshape=(None,))
                    if save_truth:
                        truth_dataset = f.create_dataset(name='truth_t',
                                                         data=truth_t,
                                                         maxshape=(truth_t.shape[0], None))
                    target_dataset = f.create_dataset(name='target',
                                                      data=target,
                                                      maxshape=(None,))
                    noise_dataset = f.create_dataset(name='noise',
                                                     data=noise,
                                                     maxshape=(None,))
                    if compute_segsnr:
                        segsnr_dataset = f.create_dataset(name='segsnr',
                                                          data=segsnr,
                                                          maxshape=(None,))
                    truth_elems = truth_t.shape[0]
                else:
                    mixture_dataset.resize(sample_offset + samples, axis=0)
                    mixture_dataset[sample_offset:] = mixture
                    if save_truth:
                        truth_dataset.resize(sample_offset + samples, axis=1)
                        truth_dataset[:, sample_offset:] = truth_t
                    target_dataset.resize(sample_offset + samples, axis=0)
                    target_dataset[sample_offset:] = target
                    noise_dataset.resize(sample_offset + samples, axis=0)
                    noise_dataset[sample_offset:] = noise
                    if compute_segsnr:
                        segsnr_dataset.resize(sample_offset + samples, axis=0)
                        segsnr_dataset[sample_offset:] = segsnr
                sample_offset += samples

                for idx, val in enumerate(m):
                    mixdb_out['mixtures'][val] = mixdb_tmp['mixtures'][idx]
                for idx in range(mixdb_out['num_classes']):
                    mixdb_out['class_count'][idx] += mixdb_tmp['class_count'][idx]
                chunk_offset += chunk_size
            f.attrs['mixdb'] = json.dumps(mixdb_out)
        progress.close()

        logger.info('Wrote {}'.format(output_name))
        duration = total_samples / sonusai.mixture.sample_rate
        logger.info('')
        logger.info('Duration: {}'.format(seconds_to_hms(seconds=duration)))
        logger.info('mixture:  {}'.format(human_readable_size(sample_offset * 2, 1)))
        if save_truth:
            logger.info('truth_t:  {}'.format(human_readable_size(sample_offset * truth_elems * 4, 1)))
        logger.info('target:   {}'.format(human_readable_size(sample_offset * 2, 1)))
        logger.info('noise:    {}'.format(human_readable_size(sample_offset * 2, 1)))
        if compute_segsnr:
            logger.info('segsnr:   {}'.format(human_readable_size(sample_offset * 4, 1)))

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()


if __name__ == '__main__':
    main()



import numpy as np

from sonusai import logger
from sonusai.mixture import apply_augmentation
from sonusai.mixture import get_next_noise
from sonusai.mixture import get_noise_audio_from_db
from sonusai.mixture import get_target_audio_from_db


def process_mixture_audio(i_frame_offset: int,
                          i_sample_offset: int,
                          o_frame_offset: int,
                          mixdb: dict,
                          mixture_record: dict,
                          target_audios: list,
                          noise_audios: list) -> (np.ndarray, np.ndarray):
    if mixture_record['samples'] % mixdb['frame_size'] != 0:
        logger.error('Number of samples in mixture is not a multiple of {}'.format(mixdb['frame_size']))
        exit()
    target_file_index = mixture_record['target_file_index']
    target_augmentation = mixdb['target_augmentations'][mixture_record['target_augmentation_index']]
    target_audio = apply_augmentation(audio_in=get_target_audio_from_db(target_audios, target_file_index),
                                      augmentation=target_augmentation,
                                      length_common_denominator=mixdb['feature_step_samples'],
                                      dither=mixdb['dither'])
    if len(target_audio) != mixture_record['samples']:
        logger.error('Number of samples in target does not match database')
        exit()
    noise_file_index = mixture_record['noise_file_index']
    noise_augmentation_index = mixture_record['noise_augmentation_index']
    noise_audio, _ = get_next_noise(offset_in=mixture_record['noise_offset'],
                                    length=mixture_record['samples'],
                                    audio_in=get_noise_audio_from_db(noise_audios,
                                                                     noise_file_index,
                                                                     noise_augmentation_index))
    mixture_record['i_sample_offset'] = i_sample_offset
    mixture_record['i_frame_offset'] = i_frame_offset
    mixture_record['o_frame_offset'] = o_frame_offset

    target_audio = np.int16(np.single(target_audio) * mixture_record['target_snr_gain'])
    noise_audio = np.int16(np.single(noise_audio) * mixture_record['noise_snr_gain'])

    return noise_audio, target_audio



from typing import List

import numpy as np


def format_confusion_matrix(cm: np.ndarray, labels: List[str] = None) -> str:
    # print confusion matrix from float variable cm and optional labels list
    result = '{}\n'.format(cm.astype(int))
    return result



def seconds_to_hms(seconds: float) -> str:
    h = int(seconds / 3600)
    s = seconds - h * 3600
    m = int(s / 60)
    s = int(seconds - h * 3600 - m * 60)
    return '{:d}:{:02d}:{:02d} (H:MM:SS)'.format(h, m, s)



import logging
from os import getcwd
from os.path import dirname

__version__ = '0.3.17'
basedir = dirname(__file__)

# create logger
logger = logging.getLogger('sonusai')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(message)s')
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


# create file handler
def create_file_handler(filename: str):
    fh = logging.FileHandler(filename=filename, mode='w')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


# update console handler
def update_console_handler(verbose: bool):
    if not verbose:
        logger.removeHandler(console_handler)
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)


# write initial log message
def initial_log_messages(name: str):
    from datetime import datetime
    from getpass import getuser
    from socket import gethostname

    logger.info('SonusAI {} {}'.format(name, __version__))
    logger.debug('Host:      {}'.format(gethostname()))
    logger.debug('User:      {}'.format(getuser()))
    logger.debug('Directory: {}'.format(getcwd()))
    logger.debug('Date:      {}'.format(datetime.now()))
    logger.debug('')



import numpy as np

from sonusai import logger
from sonusai.mixture import apply_augmentation
from sonusai.mixture import read_audio


def build_noise_audio_db(mixdb: dict) -> list:
    db = []
    for file_index in sorted(list(set([sub['noise_file_index'] for sub in mixdb['mixtures']]))):
        audio_in = read_audio(name=mixdb['noises'][file_index]['name'], dither=mixdb['dither'])
        for augmentation_index in sorted(list(set([sub['noise_augmentation_index'] for sub in mixdb['mixtures']]))):
            db.append({
                'file':         file_index,
                'augmentation': augmentation_index,
                'audio':        apply_augmentation(audio_in=audio_in,
                                                   augmentation=mixdb['noise_augmentations'][augmentation_index],
                                                   length_common_denominator=1,
                                                   dither=mixdb['dither'])
            })
    return db


def get_noise_audio_from_db(db: list, file_index: int, augmentation_index: int) -> np.ndarray:
    for record in db:
        if record['file'] == file_index and record['augmentation'] == augmentation_index:
            return record['audio']

    logger.error(
        'Could not find noise_file_index {} and noise_augmentation_index {} in noise database'.format(file_index,
                                                                                                      augmentation_index))
    exit()


def build_target_audio_db(mixdb: dict) -> list:
    db = []
    for file_index in sorted(list(set([sub['target_file_index'] for sub in mixdb['mixtures']]))):
        db.append({
            'file':  file_index,
            'audio': read_audio(name=mixdb['targets'][file_index]['name'], dither=mixdb['dither'])
        })
    return db


def get_target_audio_from_db(db: list, file_index: int) -> np.ndarray:
    for record in db:
        if record['file'] == file_index:
            return record['audio']

    logger.error('Could not find target_file_index {} in target database'.format(file_index))
    exit()
